// DS MMU emulator
// Mic, 2005

#include <windows.h>
#include <commctrl.h>
#include <string.h>
#include <bass.h>
#include <mmsystem.h>
#include <mmreg.h>
#include <stdio.h>
#include "zombie.h"

#define APU_FREQ 32000
#define APU_CHN 1

#define APU_VERSION 18



typedef struct {
	int hz, len;
	bool stereo;
	short *buf;
	int pos[2],offset,datapos,cnt;
} tag_pcm;


typedef struct {
	unsigned int fmt,sad,pos,delta,len,looptype,looppoint,n,freq;
	short vol;
	bool enabled;
} tag_sndchan;


HINSTANCE		hInst;
HWAVEOUT		hWav;
HANDLE			hout;
WAVEFORMATEX	waveFormat;
WAVEHDR			waveHdr[2];
DWORD			oldVolume,dummy;
HSTREAM			hStream;
unsigned char	*IOREGS,*IOREGS_ARM7,*VRAM;
unsigned int	*cycle=NULL,cycles;
unsigned int	cpu_snd;
short			*stream[2];
zombie_callback zombie;
char			apuName[] = "DS APU";
char			waveOutInitError[] = "Failed to initialize vaveOut device";
char			memoryError[] = "Unable to allocate sufficient memory";
char			*lpszError;
char			szbuffer[256];
bool			audio_done;
int				buffer_played;
tag_pcm			pcm;
tag_sndchan		channels[16];
unsigned char	**arm7_pageTable;
unsigned int	*arm7_pageMask;
bool			terminate;
short			sndbias;

bool __stdcall DllMain(HINSTANCE hDLL,DWORD reason,LPVOID unused)
{
	if (reason==DLL_PROCESS_ATTACH)
		hInst = hDLL;

	return true;
}



///////////////////////////////////////////////////////////////////////////////////////

void apu_mix();


DWORD CALLBACK stream_proc(HSTREAM handle, void *buffer, DWORD length, DWORD user) {
	pcm.len = length>>1;
	stream[0] = (short*)buffer;
	buffer_played = 0;
	apu_mix();

	return length;
}



bool __stdcall apu_init(zombie_callback commlink) {
	int i,j;

	zombie = commlink;

	audio_done = false;

	pcm.hz = APU_FREQ;
	pcm.len = APU_FREQ/4; /* PCM buffer length */
	pcm.stereo = false;

	if (BASS_Init(1,APU_FREQ,BASS_DEVICE_MONO,0,0) == FALSE)
		return false;

	if ((hStream = BASS_StreamCreate(APU_FREQ,1,0,(STREAMPROC*)stream_proc,NULL)) == 0)
		return false;

	hout = GetStdHandle(STD_OUTPUT_HANDLE);
	return true;

}



bool __stdcall apu_reset() {
	int i;

	VRAM = (unsigned char*)zombie(ASK,GPU_MEMORY,(void*)0L,(void*)0L);
	if ((int)VRAM == 0L)
		return false;

	IOREGS		= &VRAM[IOREGS_OFFSET];
	IOREGS_ARM7 = &VRAM[IOREGS_ARM7_OFFSET];

	IOREGS_ARM7[0x501] = 0;
	
	for (i=0; i<16; i++) {
		channels[i].pos = 0;
		channels[i].enabled = false;
		channels[i].delta = 0;

		*(unsigned short*)(IOREGS_ARM7 + 0x40A + (i<<4)) = 0;
	}

	terminate = false;
	buffer_played = 0;
	sndbias = 0;

	BASS_ChannelStop(hStream);
	BASS_StreamPlay(hStream,TRUE,0);

	return true;
}


void __stdcall apu_close() {
	terminate = true;

	BASS_ChannelStop(hStream);
	BASS_StreamFree(hStream);
	BASS_Free();
}



int __stdcall apu_config(HWND hwnd) {
	return 0; 
}


void apu_mix() {
	int i,j;
	int sample;
	unsigned int pos,addr;
	short smp;

	if (terminate) return;

	if ((IOREGS_ARM7[0x501]&0x80)==0) {	/* Check sound master enable bit */
		memset(stream[buffer_played],0,pcm.len*2);
		return;

	} else {
		for (i=0; i<pcm.len; i++) {
			sample = 0;
			for (j=0; j<16; j++) {
				if (channels[j].enabled) {

					/* PCM8 */
					if (channels[j].fmt == 0) {
						addr = channels[j].sad + (channels[j].pos >> 9);
						if (arm7_pageTable[addr>>20]) {
							smp = *(char*)(arm7_pageTable[addr>>20] + (addr & arm7_pageMask[addr>>20]));
							sample += (smp*channels[j].vol*2);
						}

					/* PCM16 */
					} else if (channels[j].fmt == 1) {
						addr = channels[j].sad + ((channels[j].pos >> 9)<<1);
						if (arm7_pageTable[addr>>20]) {
							smp = *(short*)(arm7_pageTable[addr>>20] + (addr & arm7_pageMask[addr>>20]));
							sample += (smp*channels[j].vol)>>7;
						}
					}

					channels[j].pos += channels[j].delta;
					if ((channels[j].pos>>9) >= channels[j].len) {
						if (channels[j].looptype == 1) {
							channels[j].pos = channels[j].looppoint + (((channels[j].pos>>9) - channels[j].len)<<9);
						} else {
							channels[j].pos = 0;
							channels[j].enabled = false;
							IOREGS_ARM7[0x403 + (j<<4)] &= 0x7F;
						}
					}
				}
			}

			/* Apply bias and saturation */
			sample += sndbias;
			if (sample > 32767) sample = 32767;
			else if (sample < -32768) sample = -32768;

			stream[buffer_played][i] = sample;
		}
	}
}


void __stdcall apu_write(unsigned int offset, unsigned int data) {
	int chn,reg;
	int freq;

	offset &= 0xFFF;

	if (offset<0x500) {
		chn = (offset>>4)&0xF;
		switch (offset&0xF) {
			case 0:				// SOUND*CNT
			case 1:
			case 2:
			case 3:
				if (IOREGS_ARM7[offset+3]&0x80) {
					if (!channels[chn].enabled) {
						channels[chn].pos = channels[chn].looppoint = (*(unsigned short*)(IOREGS_ARM7 + 0x40A + (chn<<4)) << 9);
						channels[chn].looptype = (((*(int*)(IOREGS_ARM7 + 0x400 + (chn<<4)))>>27)&3); 
						channels[chn].enabled = true;
						channels[chn].fmt = (((*(int*)(IOREGS_ARM7 + 0x400 + (chn<<4)))>>29)&3);
						channels[chn].vol = ((IOREGS_ARM7[0x500]&0x7f));
						channels[chn].vol *= ((IOREGS_ARM7[0x400 + (chn<<4)]&0x7f));
						channels[chn].vol >>= 7;
						channels[chn].len = channels[chn].n * ((channels[chn].fmt==0)?4:2);
						/*wsprintf(szbuffer,"Channel %d volume=%d, pos=%d, lt=%d, fmt=%d, len=%d\n",
							chn,channels[chn].vol,channels[chn].pos,channels[chn].looptype,
							channels[chn].fmt,channels[chn].len);
						WriteConsole(hout,szbuffer,lstrlen(szbuffer),(unsigned long*)&dummy,NULL);*/
					}
				} else {
					channels[chn].enabled = false;
				}
				break;
			
			case 4:				// SOUND*SAD
				channels[chn].sad = *(unsigned int*)(IOREGS_ARM7 + 0x404 + (chn<<4));
				channels[chn].sad &= 0x07FFFFFF;
				/*wsprintf(szbuffer,"Setting source address %X for channel %d\n",
					channels[chn].sad,chn);
				WriteConsole(hout,szbuffer,lstrlen(szbuffer),(unsigned long*)&dummy,NULL);*/
				break;

			case 8:				// SOUND*TMR
				freq = 16777216*16 / (0x10000 - *(unsigned short*)(IOREGS_ARM7 + 0x408 + (chn<<4)));
				channels[chn].delta = (freq<<5)/APU_FREQ;
				channels[chn].freq = freq;
				/*wsprintf(szbuffer,"Setting frequency %d for channel %d, delta=%d\n",
					freq>>4,chn,channels[chn].delta);
				WriteConsole(hout,szbuffer,lstrlen(szbuffer),(unsigned long*)&dummy,NULL);*/
				break;
			
			case 10:			// SOUND*PNT
				break;

			case 12:			// SOUND*LEN
				channels[chn].n = *(unsigned int*)(IOREGS_ARM7 + 0x40C + (chn<<4));
				channels[chn].n &= 0x003FFFFF;
				channels[chn].len = channels[chn].n * ((channels[chn].fmt==0)?4:2);
				/*wsprintf(szbuffer,"Channel %d volume=%d, len=%d, pos=%d, lt=%d\n",
					chn,channels[chn].vol,channels[chn].len,channels[chn].pos,channels[chn].looptype);
				WriteConsole(hout,szbuffer,lstrlen(szbuffer),(unsigned long*)&dummy,NULL);*/
				break;
		}
	}

	if (offset==0x504) {
		sndbias = *(short*)(IOREGS_ARM7 + 0x504);
		sndbias = (sndbias-0x200)*0x20;
	}

	if ((IOREGS_ARM7[0x501]&0x80)==0)	// Check sound master enable bit
		return;


	return;
}


int __stdcall apu_get(int what) {
	switch (what) {
		case APU_WRITE_CALLBACK:
			return (int)apu_write;
		case APU_MIX_CALLBACK:
			return (int)apu_mix;
		case PLUGIN_NAME:
			return (int)&apuName;
		case PLUGIN_VERSION:
			return APU_VERSION;
		case LAST_ERROR:
			return (int)lpszError;
		default:
			break;
	}
	return 0;
}


void __stdcall apu_set(int what,void* param1,void* param2) {
	switch (what) {
		case CPU_CYCLE_PTR:
			cycle = (unsigned int *)param1;
			break;

		case MMU_PAGE_TABLE:
			arm7_pageTable = (unsigned char**)param1;
			break;

		case MMU_PAGE_MASKS:
			arm7_pageMask = (unsigned int*)param1;
			break;

		case APU_PAUSE:
			if ((int)param1==1) {
				BASS_ChannelStop(hStream);
			} else {
				BASS_StreamPlay(hStream,TRUE,0);
			}

		default:
			break;
	}
}

